require 'active_model'

module FormLogic #:nodoc:
  class Base
    include ActiveModel::Model

    cattr_reader :declared_model_name
    attr_reader  :instances

    class << self
      ##
      # @param [Sting, Symbol]
      # @example 
      #   declare_model_name :user
      # This name will be use to build an [ActiveModel::Name]
      def declare_model_name(name)
        @@declared_model_name = name.to_s
      end

      ##
      # @param [String, Symbol] name of an attribute
      # @option to [Symbol] object-receiver (the name of an instance)
      # @example 
      #   property :email, :to => :user
      #
      # Actually this is just an syntax sugar over the `delegate` method
      def property(*names, to: nil)
        receiver = to || (raise ArgumentError.new ':to option must be set') 
        names.each do |name|
          class_eval { delegate(name, to: receiver) }  
        end
      end

      # @param [Symbol]
      def virtual(*names)
        names.each do |name|
          class_eval { attr_accessor(name) }
        end
      end

      def model_name
        ActiveModel::Name.new(self, nil, declared_model_name || self.name)
      end
    end

    ##
    # @example UserForm.new(User.new(params[:user]))
    def initialize(*instances)
      @instances = instances
      instances.each do |instance|
        self.class.send(:define_method, instance_to_accessor_name(instance)) { instance }
      end
    end

    ##
    # Copies vaidation errors of declared models to self
    def valid?
      super
      copy_errors_to_self!
      if self.class.method_defined?(:while_validation)
        while_validation()
      end
      !errors.present?
    end

  protected

    # copies errors from passed instancies to self
    def copy_errors_to_self!
      instances.each do |instance|
        instance.valid?
        instance.errors.each do |attribute, description|
          self.errors.add(attribute, description)
        end
      end
    end

  private

    def instance_to_accessor_name(instance)
      instance.class.name.split('::').join('_').underscore
    end
  end
end
